#include <iostream>
#include <functional>
#include <thread>
#include <mutex>
#include <zmq.hpp>
using namespace std;

static zmq::context_t ZMQ_CONTEXT(1);
static bool stopThread(false);

static std::string ipcAddress("ipc:///tmp/ipcSocketTest");

class SocketMonitor: public zmq::monitor_t
{
public:
	SocketMonitor(std::string identifier) :
			Identifier(identifier)
	{
	}

	std::string Identifier;
private:
	void on_event_disconnected(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received DISCONNECTED event" << std::endl;
	}
	void on_event_accepted(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received ACCEPTED event" << std::endl;
	}
	void on_event_connected(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received CONNECTED event" << std::endl;
	}
	void on_event_connect_delayed(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received CONNECT DELAYED event" << std::endl;
	}
	void on_event_connect_retried(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received CONNECT RETRIED event" << std::endl;
	}
	void on_event_listening(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received LISTENING event" << std::endl;
	}
	void on_event_bind_failed(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received BIND FAILED event" << std::endl;
	}
	void on_event_accept_failed(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received ACCEPT FAILED event" << std::endl;
	}
	void on_event_closed(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received CLOSED event" << std::endl;
	}
	void on_event_close_failed(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received CLOSE FAILED event" << std::endl;
	}
	void on_event_unknown(const zmq_event_t &event_, const char* addr_) override
	{
		std::cout << Identifier << " received UNKNOWN event" << std::endl;
	}
};

//Création du couple de socket DEALER
zmq::socket_t server(ZMQ_CONTEXT, ZMQ_DEALER);
zmq::socket_t client(ZMQ_CONTEXT, ZMQ_DEALER);

//Création des monitor pour les sockets
SocketMonitor clientMonitor("Client");
SocketMonitor serverMonitor("Server");

//Création de mutex pour synchroniser la lecture et l'écriture sur chaque socket
std::mutex serverMutex;
std::mutex clientMutex;

/*
 * @brief fonction qui envoie régulièrement des données lorsque c'est possible
 */
void sendSomeData(zmq::socket_t& socket, std::string textToSend, std::mutex& socketMutex)
{
	while (not stopThread)
	{
		socketMutex.lock();
		socket.send(textToSend.c_str(), textToSend.size(), ZMQ_DONTWAIT);
		socketMutex.unlock();
		std::this_thread::sleep_for(std::chrono::milliseconds(std::rand() % 1000));
	}
}

/*
 * @brief fonction qui lit régulièrement les données reçues s'il y en a.
 */
void receiveSomeData(zmq::socket_t& socket, std::string id, std::mutex& socketMutex)
{
	while (not stopThread)
	{
		zmq::message_t receivedMessage;
		socketMutex.lock();
		bool isMessageReceived = socket.recv(&receivedMessage, ZMQ_DONTWAIT);
		socketMutex.unlock();
		if (isMessageReceived)
		{
			std::string receivedString(static_cast<char*>(receivedMessage.data()), receivedMessage.size());
			std::cout << id << " received " << receivedString << std::endl;
			std::this_thread::sleep_for(std::chrono::milliseconds(100));
		}
	}
}

/*
 * @brief fonction qui démarre la surveillance d'une socket par un monitor
 */
void startMonitor(zmq::monitor_t& monitorToStart, zmq::socket_t& socketToMonitor, std::string identifier)
{
	std::string internalAddress("inproc://" + identifier + ".monitor");
	monitorToStart.monitor(socketToMonitor, internalAddress.c_str(),
	ZMQ_EVENT_ALL);
}

int main()
{
	//Démarrage de la surveillance du server
	std::thread serverMonitorThread(std::bind(startMonitor, std::ref(serverMonitor), std::ref(server), "Server"));
	//Démarrage de la surveillance du client
	std::thread clientMonitorThread(std::bind(startMonitor, std::ref(clientMonitor), std::ref(client), "Client"));

	//Connection du server
	server.bind(ipcAddress.c_str());

	//Démarrage du thread qui envoie "ici server" au client via la fonction préalablement décrite « sendSomeData »
	std::thread serverSendThread(std::bind(sendSomeData, std::ref(server), "ici server", std::ref(serverMutex)));

	//Démarrage du thread de réception des données du client via la fonction préalablement décrite « receiveSomeData »
	std::thread serverReceiveThread(std::bind(receiveSomeData, std::ref(server), "Serveur", std::ref(serverMutex)));


	//Connection du client
	client.connect(ipcAddress.c_str());

	//Démarrage du thread qui envoie "ici client" au server via la fonction préalablement décrite « sendSomeData »
	std::thread clientSendThread(std::bind(sendSomeData, std::ref(client), "ici client", std::ref(clientMutex)));

	//Démarrage du thread de réception des données du server via la fonction préalablement décrite « receiveSomeData »
	std::thread clientReceiveThread(std::bind(receiveSomeData, std::ref(client), "Client", std::ref(clientMutex)));

	std::this_thread::sleep_for(std::chrono::seconds(5));

	std::cout << "Disconnecting server" << std::endl;
	server.unbind(ipcAddress.c_str());
	std::this_thread::sleep_for(std::chrono::seconds(1));

	std::cout << "Reconnecting server" << std::endl;
	server.bind(ipcAddress.c_str());
	std::this_thread::sleep_for(std::chrono::seconds(1));

	std::cout << "Disconnecting client" << std::endl;
	client.disconnect(ipcAddress.c_str());
	std::this_thread::sleep_for(std::chrono::seconds(1));

	std::cout << "Reconnecting client" << std::endl;
	client.connect(ipcAddress.c_str());
	std::this_thread::sleep_for(std::chrono::seconds(1));

	std::cout << "End of test" << std::endl;
	stopThread = true;
	server.close();
	client.close();

	serverReceiveThread.join();
	clientReceiveThread.join();
	serverSendThread.join();
	clientSendThread.join();
	serverMonitorThread.join();
	clientMonitorThread.join();

	return 0;
}
